/*
 * name: module hunter 2.6 ver 1.1.1
 * date: 04.9.6
 *
 *  Author:
 *  2003 		madsys   Implemented original code
 * 
 *  Fix:
 *  2004.7	madsys	make it works on 2.6 kernel
 *  2004.10	CoolQ 	Added the ability of finding module on 4g/4g patch(ingo) and PAE mode
 *                                        
 *  usage: insmod mh.ko && cat /proc/showmodules
 *                                        
 */


#include <linux/config.h>

#ifdef CONFIG_SMP
#define __SMP__ 
#endif

#if CONFIG_MODVERSIONS == 1
#define MODVERSIONS
#endif

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/unistd.h>
#include <linux/string.h>
#include <linux/mm.h>

#include <linux/proc_fs.h>

#include <linux/errno.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/fixmap.h>
#include <asm/page.h>
#include <asm/types.h>

#define HARDCORE_PD 0xc041f000
#define STEPPING (PAGE_SIZE/32)
#define __VMALLOC_RESERVE	(128 << 20)

typedef enum RAM_MODE{
	MODE_NORMAL_4K,
	MODE_NORMAL_4M,
}MODE;

static MODE 	mode;
static u32 	cr3, cr4;

/*
 * valid_addr_normal_4K,valid_addr_normal_4M
 * valid_addr_pae_4K, valid_addr_pae_2M
 * do the dirty job separately
 * return 1 for valid
 * return 0 for invalid
 */

static int valid_addr_normal_4K(u32 address)
{
	u32 	pde, pte;
	u32	cr3_tmp;
	u32	pde_idx, pte_idx;
	u32	addr;
	
	if (!address)
		return 0;

	__asm__ __volatile__ ( "movl %%cr3, %0" : "=r"(cr3) );
	cr3_tmp = cr3 & 0xfffffe00;
	//cr3_tmp = HARDCORE_PD;
	pde_idx = address >> 22;
	pte_idx = (address & 0x003ff000) >> PAGE_SHIFT;
	
	pde = ((unsigned long *) __va(cr3_tmp))[pde_idx]; /* pde */
	
	if (pde & 1){

		addr = pde & PAGE_MASK;
		//address &= 0x003ff000;
		/* pte */
		pte = ((unsigned long *) __va(addr))[pte_idx];
		if (pte)
			return 1;
		else
			return 0;
	}else
		return 0;


}


static int valid_addr_normal_4M(u32 address)
{

	u32	pde;
	u32	cr3_tmp;
	u32	pde_idx;

	if(!address)
		return 0;

	__asm__ __volatile__ ( "movl %%cr3, %0" : "=r"(cr3) );
	cr3_tmp = cr3 & 0xfffffe00;
	pde_idx = address >> 22;

	pde = ((u32 *) __va(cr3_tmp))[pde_idx];
	if(pde)
		return 1;
	else
		return 0;

}


int valid_addr(u32 address)

{
	switch(mode){
		
		case MODE_NORMAL_4K:
			return valid_addr_normal_4K(address);
			break;

		case MODE_NORMAL_4M:
			return valid_addr_normal_4M(address);
			break;

		default:
			return 0;
			break;
	}
}

ssize_t showmodule_read(struct file *unused_file, char *buffer, 
			size_t len, loff_t *off)
{
	struct module *p;

	printk("\naddress                 name     size"
		"      core_addr     flags  \n\n");

	/* brute force all the possible vms */
	for (p = (struct module *)VMALLOC_START;
	     p <= (struct module*)(VMALLOC_START + VMALLOC_RESERVE - PAGE_SIZE);
	     p = (struct module *)((u32)p + STEPPING)){
		
		if (valid_addr((u32)p + (u32)&((struct module *)NULL)->name)  )

			/* the valid module struct should match some rules */
			if ((     (p->name[0]>=0x30 && p->name[0]<=0x39)
			       || (p->name[0]> 0x41 && p->name[0]<=0x7a )) 
			    && (p->core_size < 1 <<20) 
			    && (p->core_size>= sizeof(struct module))
			    && p->state <3
			    && ((u32)p->module_core > 0x02000000)
			    && ((u32)p->module_core & 0x00000fff)  == 0)

				printk("0x%p%18s   %6lu     0x%4p     %3d\n",
					p, p->name, p->core_size, 
					p->module_core, p->state);
	}

	return 0;
}

static struct file_operations showmodules_ops = {

		read:showmodule_read,
};


int init_module(void)
{

	struct proc_dir_entry *entry;
	u32	addr;
	u32	pde, pde_idx;
	/* get cr3,cr4 registers and tell the VM mode */
	
	__asm__ __volatile__ ( "movl %%cr3, %0" : "=r"(cr3) );
	__asm__ __volatile__ ( "movl %%cr4, %0" : "=r"(cr4) );

	addr = (u32)&__this_module;
	
	pde_idx = addr >> 22;
	pde = ((u32 *) __va(cr3))[pde_idx];
		
	if(pde & (1 << 7))
		mode = MODE_NORMAL_4M;
	else
		mode = MODE_NORMAL_4K;

	printk("cr3 = 0x%x\n", cr3);

	printk("VM mode is %s\n", (mode == MODE_NORMAL_4K) 
				? "normal_4k" :"normal_4M");
	
	entry = create_proc_entry("showmodules", S_IRUSR, &proc_root);
	entry->proc_fops = &showmodules_ops;

	return 0;

}


void cleanup_module()
{
	remove_proc_entry("showmodules", &proc_root);
}



MODULE_LICENSE("GPL");
MODULE_AUTHOR("madsys<at>ercist.iscas.ac.cn; qufuping<at>ercist.iscas.ac.cn");


